// FIR filters
//
// ROM init file:	REQUIRED, with 256 or 512 coefficients.  See below.
// Input bits:		18 fixed.
// Output bits:		OBITS, default 24.
// Adder bits:		ABITS, default 24.

// FIR filter decimate by 8
module firX8R8 ( clock, x_avail, x_real, x_imag, y_avail, y_real, y_imag, coefReset, coefWR, coef);

// This requires eight MifFile's.
// Maximum NTAPS is 8 * (previous and current decimation) less overhead.
// Maximum NTAPS is 2048 (or less).
	localparam ADDRBITS	= 8;		// Address bits for 18/36 X 256 rom/ram blocks
	localparam MBITS	= 18;		// multiplier bits == input bits
	
	parameter ABITS		= 24;		// adder bits
	parameter OBITS		= 24;		// output bits
	parameter NTAPS		= 1760;		// number of filter taps, even by 8
	//parameter NTAPS		= 464;		// number of filter taps, even by 8

	input clock;
	input x_avail;							// new sample is available
	input signed [MBITS-1:0] x_real;		// x is the sample input
	input signed [MBITS-1:0] x_imag;
	output reg y_avail;						// new output is available
	output wire signed [OBITS-1:0] y_real;	// y is the filtered output
	output wire signed [OBITS-1:0] y_imag;
	input coefReset;
	input coefWR;
	input [MBITS*2-1:0] coef; 
	
	reg lastCoefWR;
	reg  [ADDRBITS-1:0] waddr;		// write sample memory address
	reg weA, weB, weC, weD, weE, weF, weG, weH;
	reg  signed [ABITS-1:0] Racc, Iacc;
	reg [3:0] cnt;
	wire signed [ABITS-1:0] RaccA, RaccB, RaccC, RaccD, RaccE, RaccF, RaccG, RaccH;
	wire signed [ABITS-1:0] IaccA, IaccB, IaccC, IaccD, IaccE, IaccF, IaccG, IaccH;
	reg signed [ABITS-1:0] Racc1, Racc2, Racc3, Racc4, Racc5, Racc6, Racc7;	// reducing fan out for Racc, Iacc
	reg signed [ABITS-1:0] Iacc1, Iacc2, Iacc3, Iacc4, Iacc5, Iacc6, Iacc7;
	reg [10:0] coefWrAddr;
	wire wrEnaA, wrEnaB, wrEnaC,wrEnaD, wrEnaE, wrEnaF, wrEnaG, wrEnaH;

	assign wrEnaA = coefWrAddr[2:0] == 3'd7 ?  coefWR & (~lastCoefWR) : 1'd0;
	assign wrEnaB = coefWrAddr[2:0] == 3'd6 ?  coefWR & (~lastCoefWR) : 1'd0;
	assign wrEnaC = coefWrAddr[2:0] == 3'd5 ?  coefWR & (~lastCoefWR) : 1'd0;
	assign wrEnaD = coefWrAddr[2:0] == 3'd4 ?  coefWR & (~lastCoefWR) : 1'd0;
	assign wrEnaE = coefWrAddr[2:0] == 3'd3 ?  coefWR & (~lastCoefWR) : 1'd0;
	assign wrEnaF = coefWrAddr[2:0] == 3'd2 ?  coefWR & (~lastCoefWR) : 1'd0;
	assign wrEnaG = coefWrAddr[2:0] == 3'd1 ?  coefWR & (~lastCoefWR) : 1'd0;
	assign wrEnaH = coefWrAddr[2:0] == 3'd0 ?  coefWR & (~lastCoefWR) : 1'd0;

	assign y_real = Racc[ABITS-1 -: OBITS];
	assign y_imag = Iacc[ABITS-1 -: OBITS];

	initial
	begin
		waddr = 0;
		weA = 0; weB = 0; weC = 0; weD = 0; weE = 0; weF = 0; weG = 0; weH = 0;
		cnt = 0;
		Racc1 = 0; Racc2 = 0; Racc3 = 0; Racc4 = 0; Racc5 = 0; Racc6 = 0; Racc7 = 0;
		Iacc1 = 0; Iacc2 = 0; Iacc3 = 0; Iacc4 = 0; Iacc5 = 0; Iacc6 = 0; Iacc7 = 0;
		coefWrAddr = NTAPS + 4*8;
		lastCoefWR = 0;
	end
	
	always @(posedge clock)
	begin
		lastCoefWR <= coefWR;

		if (coefReset)
			coefWrAddr <= 0;
		if (coefWR & (~lastCoefWR)) 
		begin
			coefWrAddr <= coefWrAddr + 1'd1;
		end
		if (x_avail)
			cnt <= cnt + 1'd1;
		if (cnt == 0)
			begin		// wait for the first x input
				y_avail <= 1'd0;
				if (x_avail)
				begin
					weA <= 1'd1;		// write new sample to memory
					Racc1 <= RaccA;		// save accumulators
					Iacc1 <= IaccA;
				end
			end
		if (cnt == 1)
			begin		// wait for the next x input
				weA <= 1'd0;
				if (x_avail)
				begin
					weB <= 1'd1;
					Racc2 <= Racc1 + RaccB;
					Iacc2 <= Iacc1 + IaccB;
				end
			end
		if (cnt == 2)
			begin
				weB <= 1'd0;
				if (x_avail)
				begin
					weC <= 1'd1;
					Racc3 <= Racc2 + RaccC;
					Iacc3 <= Iacc2 + IaccC;
				end
			end
		if (cnt == 3)
			begin
				weC <= 1'd0;
				if (x_avail)
				begin
					weD <= 1'd1;
					Racc4 <= Racc3 + RaccD;
					Iacc4 <= Iacc3 + IaccD;
				end
			end
		if (cnt == 4)
			begin
				weD <= 1'd0;
				if (x_avail)
				begin
					weE <= 1'd1;
					Racc5 <= Racc4 + RaccE;
					Iacc5 <= Iacc4 + IaccE;
				end
			end
		if (cnt == 5)
			begin
				weE <= 1'd0;
				if (x_avail)
				begin
					weF <= 1'd1;
					Racc6 <= Racc5 + RaccF;
					Iacc6 <= Iacc5 + IaccF;
				end
			end
		if (cnt == 6)
			begin
				weF <= 1'd0;
				if (x_avail)
				begin
					weG <= 1'd1;
					Racc7 <= Racc6 + RaccG;
					Iacc7 <= Iacc6 + IaccG;
				end
			end
		if (cnt == 7)
			begin		// wait for the last x input
				weG <= 1'd0;
				if (x_avail)
				begin
					weH <= 1'd1;
					Racc <= Racc7 + RaccH;
					Iacc <= Iacc7 + IaccH;
				end
			end
		if (cnt == 8)
			begin
				cnt <= cnt + 1'd1;
				weH <= 1'd0;
				y_avail <= 1'd1;		// output a sample
			end
		if (cnt == 9)
			begin
				waddr <= waddr + 1'd1;
				cnt <= 0;
			end
	end
	
	defparam A.MifFile = "coefL8A.mif", A.ABITS = ABITS;
	defparam B.MifFile = "coefL8B.mif", B.ABITS = ABITS;
	defparam C.MifFile = "coefL8C.mif", C.ABITS = ABITS;
	defparam D.MifFile = "coefL8D.mif", D.ABITS = ABITS;
	defparam E.MifFile = "coefL8E.mif", E.ABITS = ABITS;
	defparam F.MifFile = "coefL8F.mif", F.ABITS = ABITS;
	defparam G.MifFile = "coefL8G.mif", G.ABITS = ABITS;
	defparam H.MifFile = "coefL8H.mif", H.ABITS = ABITS;
	fir256 A (clock, waddr, weA, x_real, x_imag, RaccA, IaccA, coef, coefWrAddr[10:3], wrEnaA);
	fir256 B (clock, waddr, weB, x_real, x_imag, RaccB, IaccB, coef, coefWrAddr[10:3], wrEnaB);
	fir256 C (clock, waddr, weC, x_real, x_imag, RaccC, IaccC, coef, coefWrAddr[10:3], wrEnaC);
	fir256 D (clock, waddr, weD, x_real, x_imag, RaccD, IaccD, coef, coefWrAddr[10:3], wrEnaD);
	fir256 E (clock, waddr, weE, x_real, x_imag, RaccE, IaccE, coef, coefWrAddr[10:3], wrEnaE);
	fir256 F (clock, waddr, weF, x_real, x_imag, RaccF, IaccF, coef, coefWrAddr[10:3], wrEnaF);
	fir256 G (clock, waddr, weG, x_real, x_imag, RaccG, IaccG, coef, coefWrAddr[10:3], wrEnaG);
	fir256 H (clock, waddr, weH, x_real, x_imag, RaccH, IaccH, coef, coefWrAddr[10:3], wrEnaH);
endmodule

module fir256( clock, waddr, we, x_real, x_imag, Raccum, Iaccum, data, wraddress, wren);

// This filter waits until a new sample is written to memory at waddr.  Then
// it starts by multiplying that sample by coef[0], the next prior sample
// by coef[1], (etc.) and accumulating.  For R=8 decimation, coef[1] is the
// coeficient 8 prior to coef[0].

	localparam ADDRBITS	= 8;		// Address bits for 18/36 X 256 rom/ram blocks
	localparam MBITS	= 18;		// multiplier bits == input bits
	
	// These must be defined above
	parameter MifFile	= "xx.mif";	// ROM coefficients
	parameter ABITS		= 0;		// adder bits

	input clock;
	input [ADDRBITS-1:0] waddr;			// memory write address
	input we;							// memory write enable
	input signed [MBITS-1:0] x_real;	// sample to write
	input signed [MBITS-1:0] x_imag;
	output signed [ABITS-1:0] Raccum;
	output signed [ABITS-1:0] Iaccum;
	input [MBITS*2-1:0] data;
	input [7:0] wraddress;
	input wren;

	reg [ADDRBITS-1:0] raddr, caddr;			// read address for sample and coef
	wire [MBITS*2-1:0] q;						// I/Q sample read from memory
	reg  [MBITS*2-1:0] reg_q;
	wire signed [MBITS-1:0] q_real, q_imag;		// I/Q sample read from memory
	wire signed [2*MBITS-1:0] coef;				// coefficient read from memory
	reg  signed [MBITS-1:0] reg_coef_real;
	reg  signed [MBITS-1:0] reg_coef_imag;
	reg signed [MBITS*2-1:0] Rmult, Imult;		// multiplier result
	reg signed [MBITS*2-1:0] RmultSum, ImultSum;		// multiplier result
	reg [ADDRBITS:0] counter;					// count NTAPS samples + latency

	assign q_real = reg_q[MBITS*2-1:MBITS];
	assign q_imag = reg_q[MBITS-1:0];
	assign Raccum = RmultSum[MBITS*2-1 -: ABITS];
	assign Iaccum = ImultSum[MBITS*2-1 -: ABITS];

	defparam coefRam.MifFile = MifFile;
	firramH coefRam(clock, data, caddr, wraddress, wren, coef); // complex coefficient RAM 2*18 * 256
	//defparam rom.MifFile = MifFile;
	//firromH rom (caddr, clock, coef);	// coefficient ROM 18 X 256
	// sample RAM 36 X 256;  36 bit == 18 bits I and 18 bits Q
	firram36 ram (clock, {x_real, x_imag}, raddr, waddr, we, q);
	
	always @(posedge clock)
	begin
		if (we)		// Wait until a new sample is written to memory
			begin
				// last 4 coefficients have to be zero, so write 4 more and subtract here
				counter = wraddress; // + 4 - 4;	// count samples and pipeline latency
				raddr = waddr;		// read address -> newest sample
				caddr = 1'd0;		// start at coefficient zero
				//Rmult = 1'd0;
				//Imult = 1'd0;
				RmultSum = 0;
				ImultSum = 0;
			end
		else
			begin		// main pipeline here
				Rmult <= q_real * reg_coef_real - q_imag * reg_coef_imag;
				Imult <= q_imag * reg_coef_real + q_real * reg_coef_imag;
				if (counter < (wraddress + 2 - 4))
				begin
					RmultSum <= RmultSum + Rmult;
					ImultSum <= ImultSum + Imult;
				end
				if (counter>0)
				begin
					counter <= counter - 1'd1;
					raddr <= raddr - 1'd1;		// move to prior sample
					caddr <= caddr + 1'd1;		// move to next coefficient
					reg_q <= q;
					reg_coef_real <= coef[MBITS-1:0];
					reg_coef_imag <= coef[2*MBITS-1:MBITS];
				end
			end
	end
endmodule
